/** @file mission_node.cpp
 *  @version 4.0
 *  @date June 2020
 *
 *  @brief node of hotpoint 1.0/waypoint 1.0.
 *
 *  @Copyright (c) 2020 DJI
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */
#include <dji_osdk_ros/bus_driver.h>
#include <dji_osdk_ros/FlightTaskControl.h>
#include <dji_osdk_ros/ObtainControlAuthority.h>
#include <string>
#include <iostream>
#include <fstream>

using namespace DJI::OSDK;
using namespace DJI::OSDK::Telemetry;

// global variables
ros::ServiceClient     waypoint_upload_client;
ros::ServiceClient     waypoint_action_client;
ros::ServiceClient     flight_control_client;
ros::ServiceClient     obtain_ctrl_authority_client;
ros::Publisher         syncPub; // thing to publish to sync topic "positionSync"

sensor_msgs::NavSatFix gps_pos;
bool                   gps_home_flag = false;
ros::Subscriber        gps_pos_subscriber;
int                    droneId; // stupid but will work for now
int                    thresholdCounter = 0; // counts iterations before we check waypoints again
int                    thresholdLimit = 10; // number of counts before we check ...
std::vector<WayPointSettings> generatedWaypts; // waypoints to be generated, used to check if we're at waypoint or not
int                    waypointNumber;
int                    currentWaypoint = 0; // how many waypoints have we reached
long double            initialWaypointDistance; // distance to first waypoint
double                 baseSpeedParam;


// Utility function for
// converting degrees to radians
long double toRadians(const long double degree)
{
    // cmath library in C++
    // defines the constant
    // M_PI as the value of
    // pi accurate to 1e-30
    long double one_deg = (M_PI) / 180;
    return (one_deg * degree);
}

long double distance(long double lat1, long double long1,
                     long double lat2, long double long2)
{
    // Convert the latitudes
    // and longitudes
    // from degree to radians.
    lat1 = toRadians(lat1);
    long1 = toRadians(long1);
    lat2 = toRadians(lat2);
    long2 = toRadians(long2);

    // Haversine Formula
    long double dlong = long2 - long1;
    long double dlat = lat2 - lat1;

    long double ans = pow(sin(dlat / 2), 2) +
                          cos(lat1) * cos(lat2) *
                          pow(sin(dlong / 2), 2);

    ans = 2 * asin(sqrt(ans));

    // Radius of Earth in
    // Kilometers, R = 6371
    // Use R = 3956 for miles
    long double R = 6371;

    // Calculate the result
    ans = ans * R;

    // recalculate in meters
    ans = ans * 1000;

    return ans;
}

int countWaypoints()
{
        int waypointCounter = 0;
        while(ros::param::has("/robot_list/robot_"+to_string(droneId)+"/waypoint_" + to_string(waypointCounter))) // while the next waypoint exists
        {
                waypointCounter++;
        }
        ROS_INFO("counted %d waypoints in total", waypointCounter);
        return waypointCounter;
}


void gpsPosCallback(const sensor_msgs::NavSatFix::ConstPtr& msg)
{
  gps_pos = *msg;

  if (!gps_home_flag)
  {
    int generatedWaypointNumber = static_cast<int>(generatedWaypts.size());
    if (generatedWaypointNumber > 0)
    {
      initialWaypointDistance = distance(gps_pos.latitude,
        gps_pos.longitude,
        generatedWaypts[0].latitude,
        generatedWaypts[0].longitude);
      ROS_INFO("current position: (%f, %f)", static_cast<float32_t>(gps_pos.latitude), static_cast<float32_t>(gps_pos.longitude));
      ROS_INFO("first waypoint position: (%f, %f)", static_cast<float32_t>(generatedWaypts[0].latitude), static_cast<float32_t>(generatedWaypts[0].longitude));

      ROS_INFO("distance to first waypoint: %f", static_cast<float32_t>(initialWaypointDistance));
      gps_home_flag = true;
    }
  }
  else
  {
    // do checking to see if we're near a waypoint, pause if so
    if(thresholdCounter >= thresholdLimit)
    {

      thresholdCounter = 0;
      int generatedWaypointNumber = static_cast<int>(generatedWaypts.size());
      if(generatedWaypointNumber > currentWaypoint) // if we actually have waypoints left
      {
        long double tolerance = 0.5; // reached waypoint detection threshold
        long double nextWaypointDistance;
        if (currentWaypoint == 0)
        {
          nextWaypointDistance = initialWaypointDistance;
        }
        else
        {
          nextWaypointDistance =
          distance(generatedWaypts[currentWaypoint-1].latitude,
            generatedWaypts[currentWaypoint-1].longitude,
            generatedWaypts[currentWaypoint].latitude,
            generatedWaypts[currentWaypoint].longitude);
        }

          long double progressDistance =
          distance(gps_pos.latitude,
                   gps_pos.longitude,
                   generatedWaypts[currentWaypoint].latitude,
                   generatedWaypts[currentWaypoint].longitude);
          // long double reachedDistance =
          // distance(generatedWaypts[currentWaypoint].latitude,
          //          generatedWaypts[currentWaypoint].longitude,
          //          gps_pos.latitude,
          //          gps_pos.longitude);

              // now update sync topic with progress through waypoints
              std_msgs::Float32 syncUpdate;
              syncUpdate.data = static_cast<float32_t>(currentWaypoint) + static_cast<float32_t>(static_cast<float32_t>(1) - progressDistance/nextWaypointDistance);
              syncPub.publish(syncUpdate);

              if(progressDistance < tolerance)
              {
                currentWaypoint++;
                ROS_INFO("reached new waypoint! have now reached: \t%d/%d\n", currentWaypoint, generatedWaypointNumber);
              }
            }
          }
          else
          {
            thresholdCounter++;
          }
  }


}

bool runWaypointMission(int responseTimeout)
{
  ros::spinOnce();

  // Waypoint Mission : Initialization
  dji_osdk_ros::MissionWaypointTask waypointTask;
  setWaypointInitDefaults(waypointTask);

  // Waypoint Mission: Create Waypoints
  float64_t increment = 0.000001 / C_PI * 180;
  float32_t start_alt = 5;
  ROS_INFO("Creating Waypoints..\n");
  generatedWaypts = loadWaypoints();
  // generatedWaypts = createWaypoints();

  // Waypoint Mission: Upload the waypoints
  ROS_INFO("Uploading Waypoints..\n");
  uploadWaypoints(generatedWaypts, responseTimeout, waypointTask);

  // Waypoint Mission: Init mission
  ROS_INFO("Initializing Waypoint Mission..\n");
  if (initWaypointMission(waypointTask).result)
  {
    ROS_INFO("Waypoint upload command sent successfully");
  }
  else
  {
    ROS_WARN("Failed sending waypoint upload command");
    return false;
  }

  // Waypoint Mission: Start
  if (missionAction(DJI_MISSION_TYPE::WAYPOINT,
                    MISSION_ACTION::START)
        .result)
  {
    ROS_INFO("Mission start command sent successfully");
  }
  else
  {
    ROS_WARN("Failed sending mission start command");
    return false;
  }

  return true;
}

void setWaypointDefaults(WayPointSettings* wp)
{
  wp->damping         = 0.25;
  wp->yaw             = 0;
  wp->gimbalPitch     = 0;
  wp->turnMode        = 0;
  wp->hasAction       = 0;
  wp->actionTimeLimit = 100;
  wp->actionNumber    = 0;
  wp->actionRepeat    = 0;
  for (int i = 0; i < 16; ++i)
  {
    wp->commandList[i]      = 0;
    wp->commandParameter[i] = 0;
  }
}

void setWaypointInitDefaults(dji_osdk_ros::MissionWaypointTask& waypointTask)
{
  waypointTask.velocity_range     = 2;
  waypointTask.idle_velocity      = baseSpeedParam;
  waypointTask.action_on_finish   = dji_osdk_ros::MissionWaypointTask::FINISH_NO_ACTION;
  waypointTask.mission_exec_times = 1;
  waypointTask.yaw_mode		  = dji_osdk_ros::MissionWaypointTask::YAW_MODE_WAYPOINT;
  waypointTask.trace_mode	  = dji_osdk_ros::MissionWaypointTask::TRACE_COORDINATED;
  waypointTask.action_on_rc_lost  = dji_osdk_ros::MissionWaypointTask::ACTION_AUTO;
  waypointTask.gimbal_pitch_mode  = dji_osdk_ros::MissionWaypointTask::GIMBAL_PITCH_FREE;
}

std::vector<DJI::OSDK::WayPointSettings> createWaypoints()
{

  std::vector<DJI::OSDK::WayPointSettings> wpVector =
    loadWaypoints();
  return wpVector;

}

std::vector<DJI::OSDK::WayPointSettings> loadWaypoints()
{

  std::vector<double> wpt; // single waypoint
  // Let's create a vector to store our waypoints in.
  std::vector<DJI::OSDK::WayPointSettings> wp_list;
  // start_data->index = 0;
  // wp_list.push_back(*start_data);
  int i = 0;
  while(ros::param::has("/robot_list/robot_"+to_string(droneId)+"/waypoint_" + to_string(i))) // while the next waypoint exists
  {
    ros::param::get("/robot_list/robot_"+to_string(droneId)+"/waypoint_" + to_string(i), wpt);
    i++; // iterate to next parameter
    WayPointSettings wp;
    setWaypointDefaults(&wp);
    wp.index     = i;
    wp.latitude = wpt[2];
    wp.longitude = wpt[3];
    wp.altitude = wpt[4];
    wp.yaw = wpt[5];
    wp_list.push_back(wp);
  }

  return wp_list;

}

void uploadWaypoints(std::vector<DJI::OSDK::WayPointSettings>& wp_list,
                     int responseTimeout,
                     dji_osdk_ros::MissionWaypointTask& waypointTask)
{
  dji_osdk_ros::MissionWaypoint waypoint;
  double target_yaw = 0.0;
  for (std::vector<WayPointSettings>::iterator wp = wp_list.begin();
       wp != wp_list.end(); ++wp)
  {
    // target_yaw += 15.0;
    waypoint.latitude            = wp->latitude;
    waypoint.longitude           = wp->longitude;
    waypoint.altitude            = wp->altitude;
    waypoint.damping_distance    = wp->damping;
    waypoint.target_yaw          = wp->yaw;
    waypoint.target_gimbal_pitch = 0;
    waypoint.turn_mode           = wp->turnMode;
    waypoint.has_action          = 0;
    waypointTask.mission_waypoint.push_back(waypoint);
    ROS_INFO("Waypoint created at (Lat, Lon, Alt, Dmp, Yaw): %f %f %f %f %f\n ", waypoint.latitude, waypoint.longitude, waypoint.altitude, waypoint.damping_distance, target_yaw);
  }
}

ServiceAck initWaypointMission(dji_osdk_ros::MissionWaypointTask& waypointTask)
{
  dji_osdk_ros::MissionWpUpload missionWpUpload;
  missionWpUpload.request.waypoint_task = waypointTask;
  waypoint_upload_client.call(missionWpUpload);
  if (!missionWpUpload.response.result)
  {
    ROS_WARN("ack.info: set = %i id = %i", missionWpUpload.response.cmd_set,
             missionWpUpload.response.cmd_id);
    ROS_WARN("ack.data: %i", missionWpUpload.response.ack_data);
  }
  return ServiceAck(
    missionWpUpload.response.result, missionWpUpload.response.cmd_set,
    missionWpUpload.response.cmd_id, missionWpUpload.response.ack_data);
}

ServiceAck missionAction(DJI::OSDK::DJI_MISSION_TYPE type,
              DJI::OSDK::MISSION_ACTION   action)
{

  dji_osdk_ros::MissionWpAction missionWpAction;
  dji_osdk_ros::MissionHpAction missionHpAction;
  switch (type)
  {
    case DJI::OSDK::WAYPOINT:
      missionWpAction.request.action = action;
      waypoint_action_client.call(missionWpAction);
      if (!missionWpAction.response.result)
      {
        ROS_WARN("ack.info: set = %i id = %i", missionWpAction.response.cmd_set,
                 missionWpAction.response.cmd_id);
        ROS_WARN("ack.data: %i", missionWpAction.response.ack_data);
      }
      return { missionWpAction.response.result,
               missionWpAction.response.cmd_set,
               missionWpAction.response.cmd_id,
               missionWpAction.response.ack_data };
  }

}

bool takeoff()
{
  dji_osdk_ros::FlightTaskControl flightTaskControl;
  flightTaskControl.request.task = dji_osdk_ros::FlightTaskControl::Request ::TASK_TAKEOFF;
  flight_control_client.call(flightTaskControl);

  return flightTaskControl.response.result;
}

bool land()
{
  dji_osdk_ros::FlightTaskControl flightTaskControl;
  flightTaskControl.request.task = dji_osdk_ros::FlightTaskControl::Request::TASK_LAND;
  flight_control_client.call(flightTaskControl);

  return flightTaskControl.response.result;
}

int main(int argc, char** argv)
{

  //ros::Duration(7.5).sleep(); // sleep while sdk starts up// should wait for topic here


  ros::init(argc, argv, "mission_node");
  ros::NodeHandle nh;

  ros::Duration(7.5).sleep();

  ros::param::get("base_speed", baseSpeedParam);
  ros::param::get("drone_id", droneId);
  ROS_INFO("base speed set to %f", baseSpeedParam);

  // ROS stuff
  // obtain control authority?
  obtain_ctrl_authority_client = nh.serviceClient<dji_osdk_ros::ObtainControlAuthority>(
    "obtain_release_control_authority");
  waypoint_upload_client = nh.serviceClient<dji_osdk_ros::MissionWpUpload>(
    "dji_osdk_ros/mission_waypoint_upload");
  waypoint_action_client = nh.serviceClient<dji_osdk_ros::MissionWpAction>(
    "dji_osdk_ros/mission_waypoint_action");

  dji_osdk_ros::ObtainControlAuthority obtainCtrlAuthority;
  obtainCtrlAuthority.request.enable_obtain = true;
  obtain_ctrl_authority_client.call(obtainCtrlAuthority);

  syncPub = nh.advertise<std_msgs::Float32>("sync_signal", 10);

  flight_control_client = nh.serviceClient<dji_osdk_ros::FlightTaskControl>("flight_task_control");
  gps_pos_subscriber = nh.subscribe<sensor_msgs::NavSatFix>(
    "dji_osdk_ros/gps_position", 10, &gpsPosCallback);


  // Setup variables for use
  int responseTimeout = 1;

  // count number of waypoints
  waypointNumber = countWaypoints();
  runWaypointMission(responseTimeout);


  ros::spin();

  return 0;
}
